import sys
sys.path.insert(0, '.')


from .cli import check_file_system
check_file_system()
from .feature.decorators import preview, register, deregister, dropper, selector
from .feature import stl
from .feature.storage import feature_list as features
from .feature.storage import FeatureSet
from .validation.experiment import experiment_list as experiments
from .validation import Validator
from .modelling import model
from .storage.dataframe import link
from .storage import cache



from ..modelling import Model
from ..storage import cache
from .. import config
import glob
from collections import MutableSequence


class Experiment(Model):
    def __init__(self, pipeline, oofs, score, std):
        self.pipeline = pipeline
        self.model = self.pipeline.models[0].model  # TODO: test out
        self.oofs = oofs
        self.score = score
        self.std = std
        self.__name__ = f"{round(score, 3)}:exp({pipeline.__name__})"

    def __str__(self):
        string = f"({round(self.score, 5)}, std:{round(self.std, 3)}: \n\tModel: {self.pipeline.__name__})"
        return string

    def predict(self, df):
        return self.pipeline.predict(df)


class ExperimentList(MutableSequence):
    def __init__(self):
        self.experiments = []
        self.name_to_idx = dict()

    def recalc(self):
        self.experiments = []
        self.name_to_idx = dict()
        names = [obj for obj in cache.cached_objs() if obj.endswith('_exp')]
        for idx, name in enumerate(names):
            # print(idx, name)
            experiment = cache.load_obj(name)
            self.experiments.append(experiment)
            self.name_to_idx[experiment.__name__] = idx
        self.experiments.sort(key=lambda e: e.score, reverse=True)

    def __getitem__(self, item):
        """
        Implements calling to experiments by score, name and index
        :param item: str(name) or float(score), slice(score range or index range)
        :return: experiment or list of experiments
        """
        self.recalc()
        if isinstance(item, str):
            ans = [experiment for experiment in self.experiments if experiment.__name__.count(item) > 0]
        elif isinstance(item, float):
            mul = 10 ** len(str(item).split('.')[1])
            ans = [experiment for experiment in self.experiments if int(experiment.score * mul) == int(item * mul)]
        elif isinstance(item, int):
            return self.experiments[item]
        elif isinstance(item, slice):
            if type(item.start) == float:
                ans = [experiment for experiment in self.experiments if
                       item.start <= experiment.score and experiment.score < item.stop]
            else:
                return self.experiments[item]
        else:
            raise TypeError("Item must be of str, number or slice type")

        if len(ans) > 1:
            return ans
        elif len(ans) == 1:
            return ans[0]
        else:
            return

    def __repr__(self):
        self.recalc()
        string = "Experiments: [\n" + '\n'.join([experiment.__str__() for experiment in self.experiments]) + '\n]'
        return string

    def __delitem__(self, key):
        raise AttributeError('This object is read-only')

    def __setitem__(self, key, value):
        raise AttributeError('This object is read-only')

    def insert(self, key, value):
        raise AttributeError('This object is read-only')

    def register(self, experiment):
        cache.cache_obj(experiment, experiment.__name__ + '_exp')

    def __len__(self):
        self.recalc()
        return len(self.experiments)


experiment_list = ExperimentList()



import inspect
from .. import config
import os


def get_source(function):
    """
    Gets source code of a function. Ignores decorators starting with @.
    :param function: function
    :return: source code
    """
    src = inspect.getsource(function)
    if src[0] == '@':
        src = src[src.find('\n') + 1:]
    return src


def source_path(function):
    return config.source_path + f"{function.__name__}.py"


def save_source(function):
    path = source_path(function)
    with open(path, 'w') as f:
        f.write(get_source(function))


def source_is_saved(function):
    return os.path.exists(source_path(function))


def load_source(function):
    if not source_is_saved(function):
        raise FileNotFoundError
    with open(source_path(function), 'r') as f:
        source = f.read()
    return source


def matches_cache(function):
    return get_source(function) == load_source((function))



from .. import config
import numpy as np

class BaseSplitter:
    def __init__(self, y, n_folds=5, seed=config.seed):
        self.y = y
        self.n_folds = n_folds
        self.seed = seed
        
    @property
    def split(self):
        yield from self._split()
            
    @property
    def size(self):
        return self.n_folds
            
        
from sklearn.model_selection import StratifiedKFold
class SKF(BaseSplitter):
    def _split(self):
        for idx_train, idx_test in StratifiedKFold(
            self.n_folds, 
            shuffle=True, 
            random_state=self.seed
        ).split(np.zeros(len(self.y)), self.y):
            yield {'train': idx_train, 'test': idx_test}
            
        
        
from sklearn.model_selection import KFold
class KF(BaseSplitter):
    def _split(self):
        for idx_train, idx_test in KFold(
            self.n_folds, 
            shuffle=True, 
            random_state=self.seed
        ).split(np.zeros(len(self.y)), self.y):
            yield {'train': idx_train, 'test': idx_test}
            

class NM(BaseSplitter):
    def __init__(self, y, Splitter, n_folds=5, n_splits=3):
        self.y = y
        self.Splitter = Splitter
        self.n_folds = n_folds
        self.n_splits = n_splits
      
    def _split(self):
        for i in range(self.n_splits):
            yield from self.Splitter(self.y, 
                                     self.n_folds,
                                     config.seeds[i]).split
            
            
    @property
    def size(self):
        return self.n_folds * self.n_splits


from sklearn.model_selection import ShuffleSplit
from sklearn.model_selection import StratifiedShuffleSplit
class Holdout(BaseSplitter):
    def __init__(self, y, test_size=0.3, stratify=False):
        self.sz = len(y)
        self.y = y
        self.test_size = test_size
        if stratify:
            self.Splitter = StratifiedShuffleSplit
        else:
            self.Splitter = ShuffleSplit

    def _split(self):
        for idx_train, idx_test in self.Splitter(
            n_splits=1,
            test_size=self.test_size,
            random_state=self.seed
        ).split(np.zeros(len(self.y)), self.y):
            yield {'train': idx_train, 'test': idx_test}

    @property
    def size(self):
        return 1


# BUG: when using LOO split, we should score the final OOF-preds, but our current Validator will score
# each "1-sample-fold" separately using given metric instead of aggregating the predictions and scoring them all at once
# A possible solution is to set a flag "score_oofs=True" for LOO split and check it inside of Validator
from sklearn.model_selection import LeaveOneOut
class LOO(BaseSplitter):
    def __init__(self):
        pass

    def _split(self):
        for idx_train, idx_test in LeaveOneOut().split(np.zeros(len(self.y)), self.y):
            yield {'train': idx_train, 'test': idx_test}


from sklearn.cluster import KMeans
import numpy as np


try:
    from libKMCUDA import kmeans_cuda
except:
    pass


class KMeansFeaturizer:
    """Transforms numeric data into k-means cluster memberships.

    This transformer runs k-means on the input data and converts each data point
    into the id of the closest cluster. If a target variable is present, it is
    scaled and included as input to k-means in order to derive clusters that
    obey the classification boundary as well as group similar points together.

    Parameters
    ----------
    k: integer, optional, default 100
        The number of clusters to group data into.

    target_scale: float, [0, infty], optional, default 5.0
        The scaling factor for the target variable. Set this to zero to ignore
        the target. For classification problems, larger `target_scale` values
        will produce clusters that better respect the class boundary.

    random_state : integer or numpy.RandomState, optional
        This is passed to k-means as the generator used to initialize the
        kmeans centers. If an integer is given, it fixes the seed. Defaults to
        the global numpy random number generator.

    Attributes
    ----------
    cluster_centers_ : array, [k, n_features]
        Coordinates of cluster centers. n_features does count the target column.

    References
    ----------
    This implementation is partially taken from:
        "Feature Engineering for Machine Learning by Alice Zheng and Amanda Casari (O’Reilly).
         Copyright 2018 Alice Zheng and Amanda Casari, 978-1-491-95324-2."
    """

    def __init__(self, k=100, target_scale=5.0, random_state=1337, n_jobs=None):
        self.k = k
        self.target_scale = target_scale
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        """Runs k-means on the input data and find centroids.

        If no target is given (`y` is None) then run vanilla k-means on input
        `X`.

        If target `y` is given, then include the target (weighted by
        `target_scale`) as an extra dimension for k-means clustering. In this
        case, run k-means twice, first with the target, then an extra iteration
        without.

        After fitting, the attribute `cluster_centers_` are set to the k-means
        centroids in the input space represented by `X`.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_data_points, n_features)

        y : vector of length n_data_points, optional, default None
            If provided, will be weighted with `target_scale` and included in
            k-means clustering as hint.
        """
        if y is None:
            # No target variable, just do plain k-means
            try:
                clusters, assignments = kmeans_cuda(X.astype(np.float32),
                                                    clusters=self.k,
                                                    seed=self.random_state
                                                    )
                km_model = KMeans(n_clusters=self.k)
                km_model.cluster_centers_ = clusters
            except:
                km_model = KMeans(n_clusters=self.k,
                                  n_init=20,
                                  random_state=self.random_state,
                                  n_jobs=self.n_jobs
                                  )
                km_model.fit(X)

            self.km_model_ = km_model
            self.cluster_centers_ = km_model.cluster_centers_
            return self

        # There is target information. Apply appropriate scaling and include
        # into input data to k-means
        data_with_target = np.hstack((X, y[:, np.newaxis] * self.target_scale))

        # Build a pre-training k-means model on data and target
        try:
            clusters, assignments = kmeans_cuda(data_with_target.astype(np.float32),
                                                clusters=self.k,
                                                seed=self.random_state
                                                )
            km_model_pretrain = KMeans(n_clusters=self.k)
            km_model_pretrain.cluster_centers_ = clusters
        except:
            km_model_pretrain = KMeans(n_clusters=self.k,
                                       n_init=20,
                                       random_state=self.random_state,
                                       n_jobs=self.n_jobs
                                       )
            km_model_pretrain.fit(data_with_target)

        # Run k-means a second time to get the clusters in the original space
        # without target info. Initialize using centroids found in pre-training.
        # Go through a single iteration of cluster assignment and centroid
        # recomputation.
        try:
            clusters, assignments = kmeans_cuda(X.astype(np.float32),
                                                init=km_model_pretrain.cluster_centers_[:, :-1],
                                                clusters=self.k,
                                                seed=self.random_state
                                                )
            km_model = KMeans(n_clusters=self.k)
            km_model.cluster_centers_ = clusters
        except:
            km_model = KMeans(n_clusters=self.k,
                              init=km_model_pretrain.cluster_centers_[:, :-1],
                              n_init=1,
                              max_iter=1,
                              n_jobs=self.n_jobs
                              )
            km_model.fit(X)

        self.km_model = km_model
        self.cluster_centers_ = km_model.cluster_centers_
        return self

    def transform(self, X, y=None):
        """Outputs the closest cluster id for each input data point.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_data_points, n_features)

        y : vector of length n_data_points, optional, default None
            Target vector is ignored even if provided.

        Returns
        -------
        cluster_ids : array, shape[n_data_points,1]
        """
        clusters = self.km_model.predict(X)
        return clusters.reshape(-1, 1)

    def fit_transform(self, X, y=None):
        """Runs fit followed by transform.
        """
        self.fit(X, y)
        return self.transform(X, y)


from ..modelling import *


class BinaryClassifierMixin(Model):
    def predict(self, X, **kwargs):
        return self.predict_proba(X, **kwargs)[:, 1]


from xgboost import XGBClassifier as _XGBC
class XGBClassifier(_XGBC, BinaryClassifierMixin):
    short_name = 'xgb'
    tracked_params = [
        'base_score',
        'booster',
        'colsample_bylevel',
        'colsample_bytree',
        'gamma',
        'learning_rate',
        'max_delta_step',
        'max_depth',
        'min_child_weight',
        'missing',
        'n_estimators',
        'objective',
        'reg_alpha',
        'reg_lambda',
        'scale_pos_weight',
        'subsample'
    ]


from lightgbm import LGBMClassifier as _LGBMC
class LGBMClassifier(_LGBMC, BinaryClassifierMixin):
    short_name = 'lgb'
    tracked_params = [
        'class_weight',
        'boost',
        'min_data_in_leaf',
        'bagging_freq',
        'subsample_freq',
        'bagging_fraction',
        'learning_rate',
        'reg_lambda',
        'feature_fraction',
        'min_sum_hessian_in_leaf',
        'boosting_type',
        'subsample_for_bin',
        'reg_alpha',
        'tree_learner',
        'colsample_bytree',
        'min_child_weight',
        'min_child_samples',
        'subsample',
        'max_depth',
        'boost_from_average',
        'num_leaves',
        'objective',
        'n_estimators',
        'min_split_gain',
    ]


from catboost import CatBoostClassifier as _CBC
class CatBoostClassifier(_CBC, BinaryClassifierMixin):
    short_name = 'cb'
    tracked_params = [
        'iterations',
        'learning_rate',
        'l2_leaf_reg',
        'bootstrap_type',
        'bagging_temperature',
        'depth',
        'one_hot_max_size',
        'leaf_estimation_method',
        'nan_mode',
        'feature_border_type',
        'border_count',
        'max_depth',
        'colsample_bylevel',
        'eval_metric'
    ]


from sklearn.ensemble import RandomForestClassifier as _RFC
class RandomForestClassifier(_RFC, BinaryClassifierMixin):
    short_name = 'rf'
    tracked_params = [
        'bootstrap',
        'class_weight',
        'criterion',
        'max_depth',
        'max_features',
        'max_leaf_nodes',
        'min_impurity_decrease',
        'min_impurity_split',
        'min_samples_leaf',
        'min_samples_split',
        'min_weight_fraction_leaf',
        'n_estimators',
        'oob_score',
    ]


from sklearn.linear_model import LogisticRegression as _LR
class LogisticRegression(_LR, BinaryClassifierMixin):
    short_name = 'lr'
    tracked_params = [
        'C',
        'penalty',
        'fit_intercept',
        'class_weight',
    ]



from . import classification
from . import binary_classification
from . import regression
from .. import modelling
# import forge
#
# exclude_types = [model_class for model_class in modelling.__dict__.values()
#                  if isinstance(model_class, type) and modelling.Model in model_class.__bases__] + [modelling.Model]
#
# model_classes = []
# for module in (classification, binary_classification, regression):
#     model_classes += [model_class for model_class in module.__dict__.values()
#                       if isinstance(model_class, type) and modelling.Model in model_class.__bases__]
#
# model_classes = set(model_classes) - set(exclude_types)
#
#
# for model_class in model_classes:
#     model_class.__init__ = forge.copy(model_class.Estimator.__init__)(modelling.Model.__init__)
#     model_class.__doc__ = model_class.Estimator.__doc__




from .storage import FeatureConstructor
import pandas as pd
import numpy as np
from ..storage.dataframe import DataFrame as KTDF
from ..zoo.cluster import KMeansFeaturizer
from ..utils import list_hash


def empty_like(df):
    if isinstance(df, KTDF):
        return KTDF(df[[]].copy(), df.train, df.encoders, df.slice_id)
    else:
        return KTDF(df[[]].copy())


identity = FeatureConstructor(lambda df: df, cache_default=False)
identity.stl = True
identity.source = "stl.identity"


def merge(dfs):
    if len(dfs) == 1:
        return dfs[0]
    return pd.concat([df.df if isinstance(df, KTDF) else df for df in dfs], axis=1)


def column_selector(columns):
    def __col_selector(df):
        return df[columns]

    return FeatureConstructor(__col_selector, cache_default=False)


def column_dropper(columns):
    def __col_dropper(df):
        return df.drop(columns, axis=1)

    return FeatureConstructor(__col_dropper, cache_default=False)


def concat(funcs):
    def __concat(df):
        return merge([func(df) for func in funcs])

    fc = FeatureConstructor(__concat, cache_default=False)
    fc.source = f"stl.concat([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"concat_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


def compose(funcs):
    def __compose(df):
        res = empty_like(df)
        for func in funcs:
            res = merge([res, func(res)])
        return res

    fc = FeatureConstructor(__compose, cache_default=False)
    fc.source = f"stl.compose([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"compose_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


# TODO: reimplement using sklearn.preprocessing.OHE
def make_ohe(cols, sep='_ohe_'):
    def __make_ohe(df):
        return pd.get_dummies(df.df[cols], prefix_sep=sep, columns=cols)

    return FeatureConstructor(__make_ohe, cache_default=False)


def target_encoding(cols, target_col, aggregation='mean', prefix='me_'):
    def __target_encoding(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                enc = df.groupby(col)[target_col].agg(aggregation)
                df.encoders['__me_' + col] = enc
            else:
                enc = df.encoders['__me_' + col]
            res[prefix + col] = df[col].map(enc)
        return res

    return FeatureConstructor(__target_encoding, cache_default=False)


def target_encode_list(cols, target_col, aggregation='mean', prefix='me_list_'):
    def __target_encode_list(df):
        res = empty_like(df)
        filler = [np.nan]
        for col in cols:
            if df.train:
                enc = (pd.DataFrame([i if type(i) == list else [i] for i in df[col].tolist()], index=df[target_col])
                       .stack()
                       .reset_index(name=col)
                       [[col, target_col]]).groupby(col).agg(aggregation)
                df.encoders['__me_list_' + col] = enc
            else:
                enc = df.encoders['__me_list_' + col]
            res[prefix + col] = df[col].apply(lambda x: enc.loc[filter(lambda i: i in enc.index, x)].values.flatten() if type(x) == list else filler)
        return res

    return FeatureConstructor(__target_encode_list, cache_default=False)


import multiprocessing
import swifter


def apply(dataframe, function, **kwargs):
    """
    Applies function to dataframe faster.
    If n_threads is in kwargs and is greater than 1, applies by multiprocessing.
    :return: same as df.apply(function)
    """
    if 'n_threads' in kwargs:
        n_threads = kwargs.pop('n_threads')
    else:
        n_threads = 1
    if n_threads == 1:
        return dataframe.swifter.apply(function, **kwargs)

    def _apply_df(args):
        df, func, num, kw = args
        return num, df.swifter.apply(func, **kw)

    pool = multiprocessing.Pool(processes=n_threads)
    result = pool.map(_apply_df, [(d, function, i, kwargs) for i, d in enumerate(np.array_split(dataframe, n_threads))])
    pool.close()
    result = sorted(result, key=lambda x: x[0])
    return pd.concat([i[1] for i in result])


def get_categorical(df):
    if df.train:
        cat_features = [col for col in df.columns if df[col].dtype == object]
        df.encoders['__cat_features'] = cat_features
    else:
        cat_features = df.encoders['__cat_features']
    return cat_features


def get_numeric(df):
    if df.train:
        num_features = [col for col in df.columns if df[col].dtype != object]
        df.encoders['__num_features'] = num_features
    else:
        num_features = df.encoders['__num_features']
    return num_features


def discretize(cols, bins, prefix='disc_'):
    def __discretize(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + '_' + col], enc = pd.cut(df[col], bins, retbins=True)
                df.encoders[f'__disc_{bins}_{col}'] = enc
            else:
                enc = df.encoders[f'__disc_{bins}_{col}']
                res[prefix + str(bins) + '_' + col] = pd.cut(df[col], enc)
        return res

    return FeatureConstructor(__discretize, cache_default=False)


def discretize_quantile(cols, bins, prefix='disc_q_'):
    def __discretize_quantile(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + '_' + col], enc = pd.qcut(df[col], bins, retbins=True)
                df.encoders[f'__disc_q_{bins}_{col}'] = enc
            else:
                enc = df.encoders[f'__disc_q_{bins}_{col}']
                res[prefix + str(bins) + '_' + col] = pd.cut(df[col], enc)
        return res

    return FeatureConstructor(__discretize_quantile, cache_default=False)


def kmeans_encoding(cols, n_clusters, target_col=None, target_importance=5.0, prefix='km_', **kwargs):
    def __kmeans_encoding(df):
        res = empty_like(df)
        res_column_name = f'{prefix}{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}'
        if df.train:
            encoder = KMeansFeaturizer(k=n_clusters, target_scale=target_importance, **kwargs)
            if target_col:
                res[res_column_name] = encoder.fit_transform(df[cols].values, df[target_col].values)
            else:
                res[res_column_name] = encoder.fit_transform(df[cols].values)
            df.encoders[f'__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}'] = encoder
        else:
            encoder = df.encoders[f'__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}']
            res[res_column_name] = encoder.transform()
        return res

    return FeatureConstructor(__kmeans_encoding, cache_default=False)


import numpy as np
import mprop

seed = 31337
np.random.seed(seed)
seeds = np.random.randint(100, size=10)

storage_path = '../storage/'
index_prefix = "__kts__index_"
preview_call = 0


@property
def feature_path(config):
    return storage_path + 'features/'


@property
def info_path(config):
    return storage_path + 'info/'


@property
def source_path(config):
    return storage_path + 'sources/'


@property
def experiment_path(config):
    return storage_path + 'experiment/'


mprop.init()
